/* GTM 15-jul-03 
 *   - relocated filtering logic to buffer scan loop to eliminate gaps in filtered data streams
 *   - improved efficiency of buffer scan loop by keeping track of index of last record sent
 *   - eliminated false OVERFLOW messages for filtered data streams
 */

/*   GTM 13-feb-02 modified to implement filtering for data broadcasting 
 *   fork_connectors.c was modified to set intab->conn_req_typ in order to
 *   implement server-side filters for station data.
 *
 *       conn_req_type        what client gets
 *       -------------        ----------------
 *            0               everything    (original, pre-change functionality)
 *            1               '00' and '  '
 *            2               '10'
 *            3               BH?  '  ' and 00 and 10
 *            4               LH?  '  ' and 00 and 10
 *            5               VH
 *            6               everything (used to be HH)
 *            7               ('00' or '  ') and not BH, LH, VH, or HH
 *            8               '10' and not BH, LH, VH, or HH
 */

#include <dcc_std.h>
#include <dcc_misc.h>
#include <time.h>
#include "forker.h"

extern char *rec_hostname;
extern int rec_port;
extern int rec_size;
extern int rec_buff;

extern int con_port[MAX_CONN_PORTS];  /* GTM 12-feb-02 redefined con_port to be an array */

void *Fork_Client(void *);

/* This block protected by RW lock ringlock */

extern pthread_rdwr_t ringlock;

struct _rectraf {
  time_t recrec;
  int intseq;
  int orgseq;
  int ident;
  char *recbuf;
};

extern struct _rectraf *rectraf;

extern int ring_point;

/* End of ringlock block */

#define MAX_CONNECTORS 64 /* Somewhere here we exceed max files */
#define PEND_DEPTH MAX_CONNECTORS

extern struct _client_table cli_table[MAX_CONNECTORS];

extern int seqno;

int timstat=0, lastrep=0;

/* ============================================================= */

int sendclrecs(struct _client_table *intab, char *dbuf) {

  int i, j, minseq, minset, docopy, idcheck;
  int num_skipped, have_data, last_seqno;
  char bufrec[512];

  if (rec_size > 512) {
    printf("Recsize > 512???/ Bug\n");
    exit(1);
  }

  /*Prepare to look for a message to send */

  minseq = seqno+10000;
  minset = -1;
  num_skipped = 0;

  /* Lock the buffer for read */
  pthread_rdwr_rlock_np(&ringlock);

  /* attempt to locate the next data record - start search according to index of last entry sent */
  i = intab->sent_seq_index;
  last_seqno = intab->sent_seq;
  /* note that it should never be necessary to scan the entire buffer due to the breaks downstream */
  for (j = 0; j < rec_buff; j++) {
     i++;
     if (i >= rec_buff) {  /* wrap back to beginning of array */
       i = 0;
     }

     /* if the id doesn't match, then this record has never been populated with data */
     /* if the sequence number is less than the last, we've hit the boundary between new and old data */
     idcheck = (i + 5) * (i + 10);
     if ( rectraf[i].ident != idcheck || rectraf[i].intseq < last_seqno ) {
       /* adjust the index to ensure we inspect this entry again the next time around, leave last_seqno */
       i--;
       break;
     }
     /* keep track of sequence number to facilitate overflow checks. */
     last_seqno = rectraf[i].intseq;

     /* if we get here, the id matches and the sequence number is greater than that of the last sent */
     /* see if the entry meets the filtering criteria */
     have_data = 0;
     switch (intab->conn_req_type) {
     case 0:  /* no filtering, let data through  */
       have_data = 1;
       break;
     case 1:
       if ( (rectraf[i].recbuf[13] == '0' && rectraf[i].recbuf[14] == '0') 
             || (rectraf[i].recbuf[13] == ' ' && rectraf[i].recbuf[14] == ' ') ) {
         have_data = 1;
       }
       break;
     case 2:
       if ( rectraf[i].recbuf[13] == '1' && rectraf[i].recbuf[14] == '0' ) {
         have_data = 1;
       }
       break;
     case 3:
       if ( (rectraf[i].recbuf[15] == 'B' && rectraf[i].recbuf[16] == 'H')
 	   && (   (rectraf[i].recbuf[13] == '0' && rectraf[i].recbuf[14] == '0')
 	       || (rectraf[i].recbuf[13] == '1' && rectraf[i].recbuf[14] == '0') 
 	       || (rectraf[i].recbuf[13] == ' ' && rectraf[i].recbuf[14] == ' ') ) ) {
         have_data = 1;
       }
       break;
     case 4: 
       if ( (rectraf[i].recbuf[15] == 'L' && rectraf[i].recbuf[16] == 'H')
 	   && (   (rectraf[i].recbuf[13] == '0' && rectraf[i].recbuf[14] == '0')
 	       || (rectraf[i].recbuf[13] == '1' && rectraf[i].recbuf[14] == '0') 
 	       || (rectraf[i].recbuf[13] == ' ' && rectraf[i].recbuf[14] == ' ') ) ) {
         have_data = 1;
       }
       break;
     case 5: 
       if (rectraf[i].recbuf[15] == 'V' && rectraf[i].recbuf[16] == 'H') {
         have_data = 1;
       }
       break;
     case 6:  /* This used to be for triggered data, which didn't make sense.  Now, let all data through  */
       have_data = 1;
       break;
     case 7:
       if ( ( (rectraf[i].recbuf[13] == '0' && rectraf[i].recbuf[14] == '0') 
              || (rectraf[i].recbuf[13] == ' ' && rectraf[i].recbuf[14] == ' ') )
 	   &&  ! (rectraf[i].recbuf[15] == 'B' && rectraf[i].recbuf[16] == 'H')
 	   &&  ! (rectraf[i].recbuf[15] == 'L' && rectraf[i].recbuf[16] == 'H')
 	   &&  ! (rectraf[i].recbuf[15] == 'V' && rectraf[i].recbuf[16] == 'H')
 	   &&  ! (rectraf[i].recbuf[15] == 'H' && rectraf[i].recbuf[16] == 'H') ) {
         have_data = 1;
       }
       break;
     case 8:
       if ( (rectraf[i].recbuf[13] == '1' && rectraf[i].recbuf[14] == '0')
 	   &&  ! (rectraf[i].recbuf[15] == 'B' && rectraf[i].recbuf[16] == 'H')
 	   &&  ! (rectraf[i].recbuf[15] == 'L' && rectraf[i].recbuf[16] == 'H')
 	   &&  ! (rectraf[i].recbuf[15] == 'V' && rectraf[i].recbuf[16] == 'H')
 	   &&  ! (rectraf[i].recbuf[15] == 'H' && rectraf[i].recbuf[16] == 'H') ) {
         have_data = 1;
       }
       break;
     case 9: 
       if (bufrec[15] == 'L' && bufrec[16] == 'O' && bufrec[17] == 'G') {
         have_data = 1;
       }
       break;
     }  /* end of switch  */
     /* now we know if this is a record we need to send */
     if (have_data == 1) {
       minset = i;
       minseq = rectraf[i].intseq;
       break;
     } else {
       num_skipped++;
     }
  }  /* end of for(; ;) */

  /* evaluate the outcome of the scan */
  docopy=0;
  if (minset >= 0) {
    int df;

    df = minseq - intab->sent_seq - num_skipped;

    if ( df != 1 ) {
      syslog(LOG_NOTICE,"[CLIENT #%d PortID %d OVERFLOW (%d records) - %s]\n", 
	 intab->conseq, intab->conn_req_type, df-1, dbuf);
/*
printf("[CLIENT #%d PortID %d OVERFLOW (%d recs, intab->sent_seq = %d,  minseq = %d, last_seqno = %d, num_skipped =%d) - %s]\n", 
         intab->conseq, intab->conn_req_type, df-1, intab->sent_seq,  minseq, last_seqno, num_skipped, dbuf);
fprintf(stderr, "[CLIENT #%d PortID %d OVERFLOW (%d recs, intab->sent_seq: %d, minseq: %d, last_seqno: %d, num_skipped: %d) - %s]\n", 
         intab->conseq, intab->conn_req_type, df-1, intab->sent_seq,  minseq, last_seqno, num_skipped, dbuf);
*/
    }
    intab->sent_seq = minseq;
    intab->sent_seq_index = i;      /* this is where we start seaching from the next time around */

    memcpy((void *) bufrec, (void *) rectraf[minset].recbuf, rec_size);
    docopy = 1;
  } else {
    intab->sent_seq = last_seqno ;  /* track cumulative records skipped to facilitate overflow detection */
    intab->sent_seq_index = i;      /* this is where we start seaching from the next time around */
  }
  
  /* release the lock on the buffer */
  pthread_rdwr_runlock_np(&ringlock);

  /* See if we've got something to move */
  if (docopy) {
    i = send(intab->client_socket, bufrec, rec_size, 0);
    if (i <= 0) {
      return(2);
    }
    intab->lastgood = time(NULL);
    intab->reccopy ++;
  } else {
      time_t nowt = time(NULL);
      if ((nowt - intab->lastgood) > 60*5) {
	if (intab->lastgood>0) {
	  memset((void *) bufrec, '0', 6);
	  memset((void *) &bufrec[6], ' ', rec_size - 6);
	  printf("[Write out keepalive record to client #%d portid %d]\n", intab->conseq, intab->conn_req_type);
	  i = send(intab->client_socket, bufrec, rec_size, 0);
	  if (i<=0) 
	    return(2);
	}
	intab->lastgood = time(NULL);
      }
    }

  if (docopy) return(0);
  return(1);

}

void *Fork_Client(void *inptr) {

  char dbuf[200], ebuf[100], tbuf[100];
  int i, optval;
  char *e;

  struct _client_table *intab = (struct _client_table *) inptr;

  struct hostent *rethostinfo;
  unsigned char *ad_data = (unsigned char *) &intab->cli_addr.sin_addr;

  /* primary client address */
  rethostinfo = gethostbyaddr(ad_data, 4, intab->cli_addr.sin_family);
  
  if (rethostinfo) {
    sprintf(dbuf,"%s [%d.%d.%d.%d]", rethostinfo->h_name, ad_data[0], ad_data[1], ad_data[2], ad_data[3]);
  }
  else {
    sprintf(dbuf,"[%d.%d.%d.%d]", ad_data[0], ad_data[1], ad_data[2], ad_data[3]);
  }
  
  syslog(LOG_NOTICE,"[CLIENT Up #%d PortId %d - %s]\n", intab->conseq, intab->conn_req_type, dbuf);

  optval = 1;
  i = setsockopt(intab->client_socket, SOL_SOCKET, SO_KEEPALIVE, (char *) &optval, sizeof(int));

  if ( i== (-1) ) {
    printf("Couldn't set keepalive on client socket (%s)\n", strerror(errno));
  }


  for (;;) {
    int a;

    a = sendclrecs(intab, dbuf);
    if (a == 0) continue;
    if (a == 2) break;
    sched_yield();
  }

  e = strerror(errno);

  ebuf[0] = '\0';
  if (strcmp(e, "Broken pipe") !=0 ) {
    sprintf(ebuf, "(%s) ", e);
  }

  sprintf(tbuf, "%ld recs", intab->reccopy);
  if (intab->reccopy == 0) {
    sprintf(tbuf,"No Traffic");
  }

  syslog(LOG_NOTICE,"[CLIENT Down #%d Portid %d %s- %s - %s - %d secs]\n",
	 intab->conseq, intab->conn_req_type, ebuf, dbuf, tbuf,
	 time(NULL)-intab->started);

  intab->client_num = -1;	/* Free the slot */
  shutdown(intab->client_socket, 2);
  close(intab->client_socket);
  intab->client_socket = -1;
  intab->reccopy = 0;
  pthread_exit(NULL);

  exit(0);
  
}
