/* Filename: forkmain.c
 * Purpose: this is the mainline for the seedforker program
 * Modifications:
 *   July 17, 2003 - version 2.2.0
 *       - Port 4006 (default) no longer provides HH data.  It was
 *        retained for compatibility reasons, but no filtering occurs.
 *       - fork_client modified to alter identification of messages to send
 *         and to eliminate false overflow messages
 */
#include <dcc_std.h>
#include <dcc_misc.h>

#include "forker.h"

void *Fork_Receiver(void *);
void *Start_Connector(void *);
void *Start_Status_Dumper(void *);

pthread_t receiver_tid;
pthread_t connector_tid;
pthread_t dumper_tid;

time_t startup;

char *bindaddress=NULL;
char *rec_hostname;
int rec_port=4002;
int rec_size=512;
int rec_buff=128;

/*  GTM 12-feb-02 redefine con_port as array.  
    NOTE:  Fork_client's sendclrecs routine uses the array index to determine
           what data to send. Look there to see how filtering logic relates to 
           this array
*/
int con_port[MAX_CONN_PORTS]={4000, 4001, 4002, 4003, 4004, 4005, 4006, 4007, 4008, 4009};
/* if -x is present on command line, num_conn_ports will be reset to 1, effectively
   disabling the ports using filtering */
int num_conn_ports = MAX_CONN_PORTS;  
                              

char *network;
char *station;

int main(int argc, char **argv) {

  extern char *optarg;
  extern int optind;
  int c, erval;
  int verbose=0, usage=0;
  char lognam[256];

  sigset_t new_mask, orig_mask;

  startup = time(NULL);

  /* Parse our command line */
  while ((c = getopt(argc, argv, "xd:e:f:g:h:i:j:k:l:p:r:b:c:B:")) != EOF)
    switch (c) {
    case 'p':
      rec_port = atoi(optarg);
      if (rec_port == 0) usage++;
      break;
    case 'r':
      rec_size = atoi(optarg);
      if (rec_size <= 0 || rec_size > 512) usage++;
      break;
    case 'b':
      rec_buff = atoi(optarg);
      if (rec_buff == 0) usage++;
      break;
    case 'c':
      con_port[0] = atoi(optarg);
      if (con_port[0] == 0) usage++;
      break;
    case 'B':
      bindaddress = strdup(optarg);
      break;
    case '?':
      usage++;
      break;
  /* GTM 14-feb-02 added the following new command line args and removed unimplemented -v arg  */
    case 'x':  /* -x  disables filtered data streams */
      num_conn_ports = 1;
      break;
    case 'd': /* override default port for 1st filtered data stream */
      con_port[1] = atoi(optarg);
      if (con_port[1] == 0) usage++;
      break;
    case 'e': /* override default port for 2nd filtered data stream */
      con_port[2] = atoi(optarg);
      if (con_port[2] == 0) usage++;
      break;
    case 'f': /* override default port for 3rd filtered data stream */
      con_port[3] = atoi(optarg);
      if (con_port[3] == 0) usage++;
      break;
    case 'g': /* override default port for 4th filtered data stream */
      con_port[4] = atoi(optarg);
      if (con_port[4] == 0) usage++;
      break;
    case 'h': /* override default port for 5th filtered data stream */
      con_port[5] = atoi(optarg);
      if (con_port[5] == 0) usage++;
      break;
    case 'i': /* override default port for 6th filtered data stream */
      con_port[6] = atoi(optarg);
      if (con_port[6] == 0) usage++;
      break;
    case 'j': /* override default port for 7th filtered data stream */
      con_port[7] = atoi(optarg);
      if (con_port[7] == 0) usage++;
      break;
    case 'k': /* override default port for 8th filtered data stream */
      con_port[8] = atoi(optarg);
      if (con_port[8] == 0) usage++;
      break;
    case 'l': /* override default port for 9th filtered data stream */
      con_port[9] = atoi(optarg);
      if (con_port[9] == 0) usage++;
      break;    
  }
  
  if (argc - optind != 3) usage++;

  if (usage) {
    fprintf(stderr, "seedforker v2.2.0 - Receives mini-seed data from seismic stations and LISS servers and\n"); 
    fprintf(stderr, "transmits the data to clients connecting to the host on certain ports, described below.\n");
    fprintf(stderr, "Usage: seedforker -B <IP/URL>  -p <port>  -r <recsize>  -b <bufsize>  -c <port> \n");
    fprintf(stderr, "         [-x] [-d <port>] [-e <port>] [-f <port>] [-g <port>] [-h <port>] [-i <port>]\n");
    fprintf(stderr, "         [-j <port>] [-k <port>]  hostname net station\n\n");
    fprintf(stderr, "Where:\n");
    fprintf(stderr, "  -B <IP/URL>      IP address or URL of the host system running seedforker. No default.\n");
    fprintf(stderr, "  -p <port>        The port that receives LISS data.  Default is 4002.\n");
    fprintf(stderr, "  -r <recsize>     The size in bytes of LISS data records.  Upper bound is 512. Default is 512. \n");
    fprintf(stderr, "  -b <bufsize>     The number of internal data buffers (size=recsize).  Default is 128.\n");
    fprintf(stderr, "  -c <port>        The connection port for requesting all data. Default is %d.\n",con_port[0]);
    fprintf(stderr, "  -d <port>        The connection port for requesting data from Location Codes 00 and '  ',\n"); 
    fprintf(stderr, "                     all Channels. Default is %d.\n",con_port[1]);
    fprintf(stderr, "  -e <port>        The connection port for requesting data from Location Code 10, all Channels.\n");
    fprintf(stderr, "                     Default is %d.\n",con_port[2]);
    fprintf(stderr, "  -f <port>        The connection port for requesting data from Location Codes 00, 10 and '  ',\n");
    fprintf(stderr, "                     Channels BH*.  Default is %d.\n",con_port[3]);
    fprintf(stderr, "  -g <port>        The connection port for requesting data from Location Codes 00, 10 and '  ',\n");
    fprintf(stderr, "                     Channels LH*.  Default is %d.\n",con_port[4]);
    fprintf(stderr, "  -h <port>        The connection port for requesting data from Channels VH*.  Default is %d.\n",con_port[5]);
    fprintf(stderr, "  -i <port>        The connection port for requesting all data.  Default is %d.\n",con_port[6]);
    fprintf(stderr, "  -j <port>        The connection port for requesting data from Location Codes 00 and '  ',\n");
    fprintf(stderr, "                     excluding Channels BH*, LH*, VH*, and HH*.  Default is %d.\n",con_port[7]);
    fprintf(stderr, "  -k <port>        The connection port for requesting data from Location Code 10, excluding\n");
    fprintf(stderr, "                     Channels BH*, LH*, VH*, and HH*.  Default is %d.\n",con_port[8]);
    fprintf(stderr, "  -l <port>        The connection port for requesting LOG Channel data.  Default is %d.\n",con_port[9]);
    fprintf(stderr, "  -x               Disables connection ports implementing filters (corresponding to -d thru -k).\n");
    fprintf(stderr, "  hostname         IP address or URL of the LISS server from which seedforker is to receive data.\n");
    fprintf(stderr, "                     No default.\n");
    fprintf(stderr, "  net              Network Code. No default.\n");
    fprintf(stderr, "  station          Station Code. No default.\n");
    exit(0);
  }
  rec_hostname = strdup(argv[optind++]);
  network = strdup(argv[optind++]);
  station = strdup(argv[optind++]);

  Upcase(network);
  Upcase(station);

  /* Set up syslog */

  sprintf(lognam,"liss_%s_%s",network,station);
  openlog(lognam, 0, LOG_DAEMON);
  setlogmask(LOG_UPTO(LOG_DEBUG));

  /* Create the various service threads */

  (void) sigfillset(&new_mask);
  (void) sigdelset(&new_mask, SIGINT);
  (void) pthread_sigmask(SIG_SETMASK, &new_mask, &orig_mask);

  receiver_tid = 0;
  erval = pthread_create(&receiver_tid, NULL, Fork_Receiver, NULL);

  if (erval) {
    perror("pthread_create");
    bombout(erval,"Could not spawn the receiver thread (%d,%d,%d)\n",
	    erval,errno,receiver_tid);
  }

  (void) pthread_sigmask(SIG_SETMASK, NULL, &orig_mask);

  (void) sigfillset(&new_mask);
  (void) sigdelset(&new_mask, SIGINT);
  (void) pthread_sigmask(SIG_SETMASK, &new_mask, &orig_mask);

  connector_tid = 0;
  erval = pthread_create(&connector_tid, NULL, Start_Connector, NULL);

  if (erval) {
    perror("pthread_create");
    bombout(erval,"Could not spawn the connector thread (%d,%d,%d)\n",
	    erval,errno,connector_tid);
  }

  (void) pthread_sigmask(SIG_SETMASK, NULL, &orig_mask);

  (void) sigfillset(&new_mask);
  (void) sigdelset(&new_mask, SIGINT);
  (void) pthread_sigmask(SIG_SETMASK, &new_mask, &orig_mask);

  dumper_tid = 0;
  erval = pthread_create(&dumper_tid, NULL, Start_Status_Dumper, NULL);

  if (erval) {
    perror("pthread_create");
    bombout(erval,"Could not spawn the status monitor thread (%d,%d,%d)\n",
	    erval,errno,dumper_tid);
  }

  (void) pthread_sigmask(SIG_SETMASK, NULL, &orig_mask);

  FOREVER {
    sleep(60);
    sched_yield();
  }

  /* This thread should not return */

  exit(0);

}
